/**
  *  \file planetscentral/dbclient/dbzset.cc
  */

#include "planetscentral/dbclient/dbzset.h"
#include "planetscentral/dbclient/dbclient.h"

DbZSet::DbZSet(DbClient& client, string_t name)
    : DbKey(client, name)
{ }

/** Set a field.
    \param name Field name
    \param value Score
    \return true iff field was new, false if updated
    ZADD key score value    int, 0 (updated), 1 (added)             zset[key][value] := score */
bool
DbZSet::setField(string_t name, int32_t value)
{
    return client.callInt(DbRequest().withString("ZADD").withString(key).withInt(value).withString(name));
}

/** Get number of elements.
    ZCARD key               int                                     zset[key].size() */
int32_t
DbZSet::size()
{
    return client.callInt(DbRequest().withString("ZCARD").withString(key));
}

/** Change score.
    \param name Field name
    \param value Score to add
    ZINCRBY key score val   double, new score                       zset[key][value] += score */
int32_t
DbZSet::addToField(string_t name, int32_t value)
{
    return client.callInt(DbRequest().withString("ZINCRBY").withString(key).withInt(value).withString(name));
}

/** Get position of a field in the ZSet (0=first).
    \param name Field name
    \param dir Ascending or descending sort
    ZRANK key value         int-or-nil                              position of value in zset, ascending score
    ZREVRANK key value      int-or-nil                              position of value in zset, descending score */
int32_t
DbZSet::indexOf(string_t name, Direction dir)
{
    return client.callInt(DbRequest()
                          .withString(dir == Ascending ? "ZRANK" : "ZREVRANK")
                          .withString(key)
                          .withString(name));
}

/** Remove field by name.
    \param name Field name
    \return true if removed, false if not there
    ZREM key value          int, 0 (not there), 1 (removed)         zset[key].erase(value) */
bool
DbZSet::removeField(string_t name)
{
    return client.callInt(DbRequest().withString("ZREM").withString(key).withString(name));
}

/** Get value of a field. If the field does not exist, returns 0.
    \param name Field name
    ZSCORE key value        double-or-nil                           zset[key][value] */
int32_t
DbZSet::get(string_t name)
{
    return client.callInt(DbRequest().withString("ZSCORE").withString(key).withString(name));
}

/** Check presence of a field. */
bool
DbZSet::hasField(string_t name)
{
    IntValue* v = client.call(DbRequest().withString("ZSCORE").withString(key).withString(name));
    bool result = (v != 0);
    delete v;
    return result;
}

/** Get range. Returns a list of field names.
    \param start Index of first item to return (inclusive)
    \param end Index of last item to return (inclusive)
    \param dir Count in ascending or descending order
    ZRANGE k s e [WITHSCORES]  list<string>, values [and scores]    slice of sorted zset, ascending score
    ZREVRANGE k s e [WITHSCORES]  list<string>, vals [and scores]   slice of sorted zset, deascending score */
void
DbZSet::getRange(int32_t start, int32_t end, Direction dir, StringList_t& out)
{
    std::auto_ptr<IntValue> val(client.call(DbRequest()
                                            .withString(dir == Ascending ? "ZRANGE" : "ZREVRANGE")
                                            .withString(key)
                                            .withInt(start)
                                            .withInt(end)));
    toStringListResult(val.get(), out);
}

/** Remove range. Removes items within an index range. Always uses ascending sort.
    \param start Index of first item to remove (inclusive)
    \param end Index of last item to remove (inclusive)
    ZREMRANGEBYRANK k s e   int, #elements removed                  ZREM k (ZRANGE k s e) */
int32_t
DbZSet::removeRange(int32_t start, int32_t end)
{
    return client.callInt(DbRequest().withString("ZREMRANGEBYRANK").withString(key).withInt(start).withInt(end));
}

/** Get a set slice. This slice describes the value (score) range low to high, inclusive. */
DbZSetSlice
DbZSet::valueRange(int32_t low, int32_t high)
{
    return DbZSetSlice(*this, low, high);
}

DbSort
DbZSet::sort()
{
    return DbSort(getClient(), getName());
}


DbZSetSlice::DbZSetSlice(DbZSet zset, int32_t low, int32_t high)
    : zset(zset),
      low(low),
      high(high)
{ }

/** Get number of elements in slice.
    ZCOUNT k i x            int                                     size of slice of sorted zset, score i..x */
int32_t
DbZSetSlice::size()
{
    return zset.getClient().callInt(DbRequest()
                                    .withString("ZCOUNT")
                                    .withString(zset.getName())
                                    .withInt(low)
                                    .withInt(high));
}

/** Remove this slice.
    \return number of elements removed
    ZREMRANGEBYSCORE k i x  int, #elements removed                  ZREM k (ZRANGEBYSCORE k i x) */
int32_t
DbZSetSlice::remove()
{
    return zset.getClient().callInt(DbRequest()
                                    .withString("ZREMRANGEBYSCORE")
                                    .withString(zset.getName())
                                    .withInt(low)
                                    .withInt(high));
}

/** Get content of this slice.
    \param out [out] Field names will be produced here
    ZRANGEBYSCORE k i x [LIMIT o c] [WITHSCORES]  list<string>...   slice of sorted zset, score i..x */
void
DbZSetSlice::get(StringList_t& out)
{
    std::auto_ptr<IntValue> val(zset.getClient().call(DbRequest()
                                                      .withString("ZRANGEBYSCORE")
                                                      .withString(zset.getName())
                                                      .withInt(low)
                                                      .withInt(high)));
    toStringListResult(val.get(), out);
}

/** Get partial content of this slice.
    \param start,end Start/end indexes, inclusive
    \param out [out] Field names will be produced here */
void
DbZSetSlice::getRange(int32_t start, int32_t end, StringList_t& out)
{
    std::auto_ptr<IntValue> val(zset.getClient().call(DbRequest()
                                                      .withString("ZRANGEBYSCORE")
                                                      .withString(zset.getName())
                                                      .withInt(low)
                                                      .withInt(high)
                                                      .withString("LIMIT")
                                                      .withInt(start)
                                                      .withInt(end)));
    toStringListResult(val.get(), out);
}
