/**
  *  \file planetscentral/host/runner.cc
  *  \brief Running External Programs
  *
  *  The main program will be using a number of file descriptors for
  *  socket I/O. In particular the file descriptors obtained by
  *  'accept' cannot necessarily be made close-on-exec reliably. To
  *  avoid these fds leaking into child processes, we isolate running
  *  child processes into a helper process:
  *
  *     Main Program <-> Helper Process <-> Actual Child
  *
  *  The helper process is started when the server starts and the
  *  environment is still sane. The main process talks to it through
  *  two pipes. It uses our standard protocol framing to pass commands
  *  and result:
  *  - a command is passed as an array-of-strings
  *  - the resulting stdout/stderr is returned as a sequence of
  *    strings, terminated by an empty string, followed by the
  *    process' exit code as a number or string.
  *
  *  Note that we'll be using multiple helper processes, which means
  *  the second helper inherits the first's pipes. This is actually a
  *  variation of the very same problem this we want to avoid, but
  *  with a much lesser impact (child processes are not connected to
  *  the network), and we know they are close-on-exec.
  *
  *  Since this module also serializes requests, it provides a nice
  *  way to limit the load of the server.
  */

#include <stdexcept>
#include <memory>
#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include "planetscentral/host/runner.h"
#include "int/array.h"

/** Constructor. Constructs idle runner. */
Runner::Runner()
{ }

/** Destructor. */
Runner::~Runner()
{ }

/** Start helper process. Call this from main() while still single-threaded. */
void
Runner::start()
{
    // create pipes
    int command_pipe[2];
    int result_pipe[2];
    if (pipe(command_pipe) < 0 || pipe(result_pipe) < 0) {
        throw std::runtime_error("Runner::start: pipe");
    }

    // avoid file handle inheritance
    fcntl(command_pipe[0], F_SETFD, FD_CLOEXEC);
    fcntl(command_pipe[1], F_SETFD, FD_CLOEXEC);
    fcntl(result_pipe[0], F_SETFD, FD_CLOEXEC);
    fcntl(result_pipe[1], F_SETFD, FD_CLOEXEC);

    // create worker
    pid_t pid = fork();
    if (pid < 0) {
        throw std::runtime_error("Runner::start: fork");
    }
    if (pid == 0) {
        // I am the child. Read commands, write results, forever.
        ::close(command_pipe[1]);
        ::close(result_pipe[0]);
        use(command_pipe[0], result_pipe[1]);
        try {
            while (!isAtEnd()) {
                serveRequest();
                flush();
            }
        }
        catch (std::exception& e) {
            sendError(e.what());
            flush();
        }
        _exit(127);
    }

    // I am the parent. Write commands, read results.
    ::close(command_pipe[0]);
    ::close(result_pipe[1]);
    use(result_pipe[0], command_pipe[1]);
}

/** Run child process.
    \param command [in] Command to run (argv)
    \param output [out] Output will be collected here
    \return Return value: 0..255 for regular exit, 1000+ for signal death, 1999 for unknown.
    \throw runtime_error if helper process encounters an API error
    (e.g. system resources exhausted) */
int32_t
Runner::run(IntDataSegment& command, string_t& output)
{
    MutexGuard guard(mtx);
    checkConnected();

    // Send command
    sendArray(command, command.size());
    flush();

    // Read result
    while (1) {
        std::auto_ptr<IntValue> val(readObject(false));
        string_t valstr(IntValue::toStringOf(val.get(), false));
        if (valstr.empty()) {
            break;
        }
        output += valstr;
    }

    // Read exit code
    std::auto_ptr<IntValue> exitCodeVal(readObject(false));
    return toIntResult(exitCodeVal.get());
}

/** Process single request. */
void
Runner::serveRequest()
{
    // Read an object. This is the command line.
    std::auto_ptr<IntValue> v(readObject(false));

    // Validate
    IntArray* a = dynamic_cast<IntArray*>(v.get());
    if (!a || a->getData()->getDimension(0) == 0) {
        sendError("Runner::serveRequest: protocol error");
        return;
    }

    // It's valid, so we can execute it. Build argv.
    std::vector<string_t> argv_str;
    for (uint32_t i = 0, n = a->getData()->getDimension(0); i < n; ++i) {
        argv_str.push_back(IntValue::toStringOf(a->getData()->content[i], false));
    }
    std::vector<const char*> argv_c;
    for (uint32_t i = 0, n = argv_str.size(); i < n; ++i) {
        argv_c.push_back(argv_str[i].c_str());
    }
    argv_c.push_back(0);

    // Pipe to communicate with child
    int fd[2];
    if (pipe(fd) < 0) {
        sendError("Runner::serveRequest: pipe failed");
        return;
    }

    // Create child
    pid_t pid = fork();
    if (pid < 0) {
        ::close(fd[0]);
        ::close(fd[1]);
        sendError("Runner::serveRequest: fork failed");
        return;
    }

    if (pid == 0) {
        // I am the child
        ::close(fd[0]);
        close();

        // Set up child's stdin/out/err
        int null = ::open("/dev/null", O_RDONLY);
        if (null >= 0) {
            dup2(null, STDIN_FILENO);
            ::close(null);
        }
        dup2(fd[1], STDOUT_FILENO);
        dup2(fd[1], STDERR_FILENO);
        ::close(fd[1]);

        // Run child
        execvp(argv_c[0], const_cast<char**>(&argv_c[0]));
        perror(argv_c[0]);
        _exit(127);
    }

    // I am the parent. Read child's stdout, and send to caller.
    ::close(fd[1]);
    char buffer[1024];
    int n;
    while ((n = ::read(fd[0], buffer, sizeof(buffer))) > 0) {
        IntStringValue sv(string_t(buffer, n));
        sendObject(&sv);
    }
    ::close(fd[0]);
    sendObject(0);

    // Child has closed its stdout. Wait for it to exit.
    int status;
    pid_t result = waitpid(pid, &status, 0);
    if (result < 0) {
        sendError("Runner::serveRequest: waitpid failed");
    } else if (WIFEXITED(status)) {
        IntIntIntValue iv(WEXITSTATUS(status));
        sendObject(&iv);
    } else if (WIFSIGNALED(status)) {
        IntIntIntValue iv(1000 + WTERMSIG(status));
        sendObject(&iv);
    } else {
        IntIntIntValue iv(1999);
        sendObject(&iv);
    }
}
