/**
  *  \file planetscentral/talk/bbout.cc
  *  \brief Talk: BBCode output
  */

#include "planetscentral/talk/bbout.h"
#include "planetscentral/talk/text.h"
#include "planetscentral/talk/bbparse.h"
#include "planetscentral/talk/bblex.h"

static const char*
getInlineTagName(TextInlineFormat minor)
{
    switch (minor) {
     case miInBold:             return "b";
     case miInItalic:           return "i";
     case miInStrikeThrough:    return "s";
     case miInUnderline:        return "u";
     case miInMonospace:        return "tt";
     default:                   return 0;
    }
}

static const char*
getInlineAttrTagName(TextInlineAttrFormat minor)
{
    switch (minor) {
     case miIAColor:   return "color";
     case miIASize:    return "size";
     case miIAFont:    return "font";
     default:          return 0;
    }
}

static const char*
getLinkTagName(TextLinkFormat minor)
{
    switch (minor) {
     case miLinkUrl:       return "url";
     case miLinkEmail:     return "email";
     case miLinkThread:    return "thread";
     case miLinkPost:      return "post";
     case miLinkGame:      return "game";
     case miLinkUser:      return "user";
     case miLinkForum:     return "forum";
     default:              return 0;
    }
}

namespace {
    class BBRenderer {
     public:
        BBRenderer(string_t& result, InlineItemSet options)
            : result(result), options(options)
            { }

        void renderChildrenPG(TextNode* n);
        void renderChildrenInline(TextNode* n, bool in_code);
        void renderAttr(TextNode* n);
        void renderLink(TextNode* n, const char* tag);
        void renderPlaintext(const string_t& str);
        void renderInline(TextNode* n, bool in_code);

        void renderPG(TextNode* n);

     private:
        string_t& result;
        InlineItemSet options;
    };
}


void
BBRenderer::renderChildrenPG(TextNode* n)
{
    for (ptr_vector<TextNode>::size_type i = 0, e = n->children.size(); i != e; ++i) {
        if (i != 0) {
            result += "\n\n";
        }
        renderPG(n->children[i]);
    }
}

void
BBRenderer::renderChildrenInline(TextNode* n, bool in_code)
{
    for (ptr_vector<TextNode>::size_type i = 0, e = n->children.size(); i != e; ++i) {
        renderInline(n->children[i], in_code);
    }
}

void
BBRenderer::renderAttr(TextNode* n)
{
    if (!n->text.empty()) {
        result += "=";
        if (n->text.find_first_of("[]") != string_t::npos) {
            result += "\"";
            result += n->text;
            result += "\"";
        } else {
            result += n->text;
        }
    }
}

void
BBRenderer::renderLink(TextNode* n, const char* tag)
{
    if (n->children.empty()
        && n->text.find_first_of("[]") == string_t::npos
        && (n->text.empty() || n->text[0] != '@')
        && (n->text.find('@') == string_t::npos || n->text.find_first_of(" \t\n\r") == string_t::npos))
    {
        /* The above condition is true if we have no children, and we are reasonably
           sure that the content of the attribute needs no escaping, so we can use
           the abbreviated format:
           - no []
           - does not start with @
           - does not contain both @ and whitespace (this would mean we risk
           generating an at-link) */
        result += "[";
        result += tag;
        result += "]";
        result += n->text;
        result += "[/";
        result += tag;
        result += "]";
    } else {
        result += "[";
        result += tag;
        renderAttr(n);
        result += "]";
        renderChildrenInline(n, false);
        result += "[/";
        result += tag;
        result += "]";
    }
}

void
BBRenderer::renderPlaintext(const string_t& str)
{
    // Look for tags
    BBLexer lex(str);
    BBLexer::Token t;
    string_t::size_type quote_end = string_t::npos;
    while ((t = lex.read()) != BBLexer::Eof) {
        if (t == BBLexer::Paragraph) {
            // Parsing this would break a paragraph, so replace by space
            result += " ";
        } else if (t == BBLexer::Text
                   || ((t == BBLexer::TagStart || t == BBLexer::TagEnd)
                       && !BBParser::isKnownTag(lex.getTag())))
        {
            // Parsing this would NOT create a tag, so render it. But we may have to quote smileys.
            string_t::size_type pos = 0;
            string_t token = lex.getTokenString();
            InlineInfo info;
            while (findInlineElement(token, pos, options, info)) {
                // Something that needs protection
                result.append(token, pos, info.start - pos);
                if (quote_end == string_t::npos) {
                    result += "[noparse]";
                }
                result.append(token, info.start, info.length);
                quote_end = result.size();

                pos = info.start + info.length;
            }
            result.append(token, pos, token.size() - pos);
        } else if (t == BBLexer::TagStart || t == BBLexer::TagEnd || t == BBLexer::AtLink || t == BBLexer::Smiley) {
            // Must be quoted
            if (quote_end == string_t::npos) {
                // We have not started protecting yet, so do it
                result += "[noparse]";
            }
            if (t == BBLexer::TagEnd && lex.getTag() == "noparse") {
                // This would end protection, so wrap a real ender inside the tag
                result += "[/[/noparse]noparse]";
                quote_end = string_t::npos;
            } else {
                // Regular case
                result += lex.getTokenString();
                quote_end = result.size();
            }
        } else {
            // Just append (can't happen)
            result += lex.getTokenString();
        }
    }
    if (quote_end != string_t::npos) {
        result.insert(quote_end, "[/noparse]");
    }
}

void
BBRenderer::renderInline(TextNode* n, bool in_code)
{
    switch (TextMajorKind(n->major)) {
     case maPlain:
        // If the text was not generated by parsing from BBCode, it might contain BBCode tags.
        // We want to keep these text, so protect them from the parser.
        // The easiest way to locate tags is to ask the parser to find them.
        // We then wrap a [noparse] around them. We try to use as few [noparse] as possible,
        // by opening one when needed, and closing it as late as possible.
        // Unfortunately, quoting within [code] is impossible, so the best we can do is to keep the code as is.
        if (in_code) {
            // We cannot quote inside this anyway.
            result += n->text;
        } else {
            renderPlaintext(n->text);
        }
        break;
     case maInline:
        if (const char* tag = getInlineTagName(TextInlineFormat(n->minor))) {
            result += "[";
            result += tag;
            result += "]";
            renderChildrenInline(n, in_code);
            result += "[/";
            result += tag;
            result += "]";
        } else {
            renderChildrenInline(n, in_code);
        }
        break;
     case maInlineAttr:
        if (const char* tag = getInlineAttrTagName(TextInlineAttrFormat(n->minor))) {
            result += "[";
            result += tag;
            renderAttr(n);
            result += "]";
            renderChildrenInline(n, in_code);
            result += "[/";
            result += tag;
            result += "]";
        } else {
            renderChildrenInline(n, in_code);
        }
        break;
     case maLink:
        if (const char* tag = getLinkTagName(TextLinkFormat(n->minor))) {
            renderLink(n, tag);
        } else {
            renderChildrenInline(n, in_code);
        }
        break;

     case maSpecial:
        switch (TextSpecialFormat(n->minor)) {
         case miSpecialImage:
            renderLink(n, "img");
            break;

         case miSpecialBreak:
            result += "[nl]";
            break;

         case miSpecialSmiley:
            result += options.contains(InlineInfo::Smiley) ? ":" : "[:";
            result += n->text;
            result += options.contains(InlineInfo::Smiley) ? ":" : ":]";
        }
        break;

     case maParagraph:
     case maGroup:
        // Should not happen
        break;
    }
}

void
BBRenderer::renderPG(TextNode* n)
{
    switch (TextMajorKind(n->major)) {
     case maGroup:
        switch (TextGroupFormat(n->minor)) {
         case miGroupQuote:
            result += "[quote";
            renderAttr(n);
            result += "]\n";
            renderChildrenPG(n);
            result += "[/quote]";
            break;
         case miGroupList:
            result += "[list";
            renderAttr(n);
            result += "]\n";
            renderChildrenPG(n);
            result += "[/list]";
            break;
         case miGroupListItem:
            result += "[*] ";
            renderChildrenPG(n);
            break;
         case miGroupRoot:
         default:
            renderChildrenPG(n);
            break;
        }
        break;
        
     case maParagraph:
        switch (TextParagraphFormat(n->minor)) {
         case miParCode:
            result += "[code";
            renderAttr(n);
            result += "]\n";
            renderChildrenInline(n, true);
            result += "[/code]";
            break;
         case miParCentered:
            result += "[center";
            renderAttr(n);
            result += "]\n";
            renderChildrenInline(n, false);
            result += "[/center]";
            break;
         default:
         case miParNormal:
            renderChildrenInline(n, false);
            break;
        }
     default:
        break;
    }
}

string_t
renderBB(TextNode* n, InlineItemSet options)
{
    string_t result;
    BBRenderer(result, options).renderChildrenPG(n);
    return result;
}
