/**
  *  \file planetscentral/talk/cmdgroup.cc
  *  \brief Talk: GROUP commands
  */

#include "planetscentral/talk/cmdgroup.h"
#include "int/args.h"
#include "int/array.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/talk/group.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/talk/talkconnection.h"
#include "planetscentral/talk/db.h"

static void
configureGroup(Group& group, IntArgBlock& args)
{
    while (args.getNumArgs() > 0) {
        args.checkArgumentCountAtLeast(2);
        string_t key(toStringResult(args.getNext()));
        string_t value(toStringResult(args.getNext()));
        if (key == "parent") {
            group.setParent(value);
        } else {
            group.header().setField(key, value);
        }
    }
}

/* @q GROUPADD id:GRID key:Str value:Str ... (Talk Command)
   Create a forum group.
   The command is followed by %key/%value pairs that configure the group, see {GROUPSET}.
   At least one pair ("name", typically) must be given.

   The %id must be a valid identifier (alphanumerics).

   Permissions: admin.

   @err 409 Already exists (the %id is already taken)
   @uses group:$GRID:header */
void
doGroupAdd(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(3);
    string_t groupId = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Group group(database_connection, groupId);
    conn.checkAdmin();
    if (group.exists()) {
        conn.sendError("409 Already exists");
    } else {
        configureGroup(group, args);
        conn.sendSuccess("OK");
    }
}

/* @q GROUPSET id:GRID [key:Str value:Str ...] (Talk Command)
   Configure forum group.
   Valid keys are the same as hash keys in {group:$GRID:header}:
   - %name ({@type Str|string}, name of group)
   - %description ({@type TalkText}, description/subtitle)
   - %parent ({@type GRID} of parent group)
   - %key ({@type Str|string}, sort key)
   - %unlisted ({@type Int}, nonzero to make this group unlistable)

   Permissions: admin.

   @err 404 Not found (%id does not correspond to a valid group)
   @uses group:$GRID:header, group:$GRID:groups */
void
doGroupSet(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(1);
    string_t groupId = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Group group(database_connection, groupId);
    conn.checkAdmin();
    if (!group.exists()) {
        conn.sendError("404 Not found");
    } else {
        configureGroup(group, args);
        conn.sendSuccess("OK");
    }
}

/* @q GROUPGET id:GRID key:Str (Talk Command)
   Get property of group.
   The %key can be a hash key as given to {GROUPSET}.

   Permissions: none.

   @uses group:$GRID:header
   @rettype GRID
   @rettype Str
   @rettype TalkText */
void
doGroupGet(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(2);
    string_t groupId = toStringResult(args.getNext());
    string_t field = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    std::auto_ptr<IntValue> result(Group(database_connection, groupId).header().getFieldRaw(field));

    // Send result
    conn.sendObject(result.get());
}

/* @q GROUPLS id:GRID (Talk Command)
   List group contents.

   Permissions: none.

   Everyone may list every group, but unlistable groups will appear empty in user contexts.
   Nonexistant groups will appear empty.

   The results will be sorted by their sort key.

   @retkey groups:StrList (list of contained groups)
   @retkey forums:IntList (list of {@talk FID|forum Ids)
   @rettype FID
   @uses group:$GRID:groups, group:$GRID:forums */
void
doGroupList(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    string_t groupId = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Group group(database_connection, groupId);
    Ptr<IntArrayData> ad(new IntArrayData());

    if (!conn.isAdmin() && group.unlisted().get()) {
        Ptr<IntArrayData> blank(new IntArrayData());
        blank->addDimension(0);
        ad->content.pushNew(makeStringValue("groups"));
        ad->content.pushNew(new IntArray(blank));
        ad->content.pushNew(makeStringValue("forums"));
        ad->content.pushNew(new IntArray(blank));
    } else {
        ad->content.pushNew(makeStringValue("groups"));
        ad->content.pushNew(group.subgroups()
                            .sort()
                            .sortLexicographical()
                            .by(Group(database_connection, "*").key())
                            .getResult());
        ad->content.pushNew(makeStringValue("forums"));
        ad->content.pushNew(group.forums()
                            .sort()
                            .sortLexicographical()
                            .by(FORUM_ROOT "*:header->key")
                            .getResult());
    }
    ad->addDimension(ad->content.size());

    // Send result
    IntArray arr(ad);
    conn.sendObject(&arr);
}

/* @q GROUPSTAT id:GRID (Talk Command)
   Get information about one group.

   Permissions: none.

   Note that this command uses the rendering configuration set with {RENDEROPTION}
   to render the group description.

   @retval TalkGroupInfo information about group
   @see GROUPMSTAT
   @uses group:$GRID:header */
void
doGroupStat(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    string_t groupId = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Group group(database_connection, groupId);
    std::auto_ptr<IntValue> result(group.describe(conn.getRenderState()));

    // Send result
    conn.sendObject(result.get());
}

/* @q GROUPMSTAT id:GRID... (Talk Command)
   Get information about multiple groups.

   Permissions: none.

   Note that this command uses the rendering configuration set with {RENDEROPTION}
   to render the group descriptions.

   @retval TalkGroupInfo[] information about groups
   @see GROUPSTAT
   @uses group:$GRID:header */
void
doGroupMStat(TalkConnection& conn, IntArgBlock& args)
{
    // Do it
    ReconnectingGuard g(database_rc);
    Ptr<IntArrayData> ad(new IntArrayData());
    while (args.getNumArgs() > 0) {
        string_t groupId = toStringResult(args.getNext());
        ad->content.pushNew(Group(database_connection, groupId).describe(conn.getRenderState()));
    }
    ad->addDimension(ad->content.size());

    // Send result
    IntArray arr(ad);
    conn.sendObject(&arr);
}
