/**
  *  \file planetscentral/talk/cmduser.cc
  *  \brief Talk: USER commands
  */

#include <algorithm>
#include <stdexcept>
#include "planetscentral/talk/cmduser.h"
#include "planetscentral/talk/newsrc.h"
#include "planetscentral/dbclient/dbsubtree.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/talkconnection.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/talk/forum.h"
#include "planetscentral/talk/topic.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/talk/user.h"
#include "planetscentral/talk/cmdforum.h"

namespace {
    class NewsrcAction {
     public:
        NewsrcAction(DbSubtree n);
        void process(int32_t messageId);
        void process(DbIntSet set);
        void save();

        void markSet();
        void markGet();
        void markClear();
        void markFind(bool value);
        void markWantId();
        bool isStopped() const;
        void sendResult(TalkConnection& conn);

     private:
        Newsrc n;

        // FIXME: every 8-bit microcontroller programmer would be proud of this
        // zoo of bools. Can we do better?
        bool get;               // we want to get the values
        bool set;               // we want to set processed items (excludes clear)
        bool clear;             // we want to clear processed items (excludes set)
        bool find;              // we want to find items (requires get)
        bool value;             // value of the items we want to find
        bool found;             // set if we found a matching item
        bool stop;              // set if we can stop processing
        bool wantId;            // set if we want the Id of the first item
        string_t result;
        int32_t firstId;
    };

    class WatchAction {
     public:
        enum What { Watch, Unwatch, MarkSeen };
        WatchAction(What w);
        void process(IntArgBlock& args, string_t user);

     private:
        const What what;
    };
}

/****************************** NewsrcAction *****************************/

NewsrcAction::NewsrcAction(DbSubtree n)
    : n(n),
      get(false),
      set(false),
      clear(false),
      find(false),
      value(false),
      found(false),
      stop(false),
      wantId(false),
      result(),
      firstId(0)
{ }

void
NewsrcAction::process(int32_t messageId)
{
    if (get) {
        bool me = n.get(messageId);
        if (find) {
            if (me == value) {
                found = true;
                if (!set && !clear) {
                    stop = true;
                }
                if (firstId == 0) {
                    firstId = messageId;
                }
            }
        } else {
            result += '0' + me;
        }
    }
    if (set) {
        n.set(messageId);
    }
    if (clear) {
        n.clear(messageId);
    }
}

void
NewsrcAction::process(DbIntSet set)
{
    // FIXME: possible optimisation: if we're just checking for any unread message
    // (find && !value && !wantId && !set && !clear), process the list backwards.
    IntList_t result;
    set.getAll(result);
    std::sort(result.begin(), result.end());
    for (IntList_t::iterator i = result.begin(); i != result.end() && !stop; ++i) {
        process(*i);
    }
}

inline void
NewsrcAction::save()
{
    n.save();
}

inline void
NewsrcAction::markSet()
{
    set = true;
    clear = false;
}

inline void
NewsrcAction::markGet()
{
    get = true;
}

inline void
NewsrcAction::markClear()
{
    set = false;
    clear = true;
}

inline void
NewsrcAction::markFind(bool value)
{
    this->get   = true;
    this->find  = true;
    this->value = value;
}

inline void
NewsrcAction::markWantId()
{
    wantId = true;
}

inline bool
NewsrcAction::isStopped() const
{
    return stop;
}

void
NewsrcAction::sendResult(TalkConnection& conn)
{
    if (get) {
        if (wantId) {
            IntIntIntValue iv(firstId);
            conn.sendObject(&iv);
        } else if (find) {
            // Searching for 1, found it --> result 1
            // Searching for 0, found it --> result 0
            // Searching for 1, none found --> result 0
            // Searching for 0, none found --> result 1
            IntIntIntValue iv(found == value);
            conn.sendObject(&iv);
        } else {
            IntStringValue sv(result);
            conn.sendObject(&sv);
        }
    } else {
        conn.sendSuccess("OK");
    }
}

static void
checkUser(TalkConnection& conn)
{
    if (conn.getUser().empty()) {
        throw std::runtime_error("401 Need user");
    }
}

/****************************** WatchAction ******************************/

WatchAction::WatchAction(What w)
    : what(w)
{ }

void
WatchAction::process(IntArgBlock& args, string_t user)
{
    // Permission checks are done when a post is submitted, so there's no need
    // to do them here. The user interface will not let users subscribe to topics
    // they are not allowed to read, so there'll not be too many database zombies.
    User u(database_connection, user);
    while (args.getNumArgs() > 0) {
        string_t keyword = strUCase(toStringResult(args.getNext()));
        if (keyword == "THREAD") {
            args.checkArgumentCountAtLeast(1);
            int32_t tid = toIntResult(args.getNext());
            Topic t(database_connection, tid);
            if (!t.exists()) {
                throw std::runtime_error("404 Not found");
            }
            if (what == Watch) {
                t.watchers().add(user);
                u.watchedTopics().add(tid);
            }
            if (what == Unwatch) {
                t.watchers().remove(user);
                u.watchedTopics().remove(tid);
            }
            u.notifiedTopics().remove(tid);
        } else if (keyword == "FORUM") {
            args.checkArgumentCountAtLeast(1);
            int32_t fid = toIntResult(args.getNext());
            Forum f(database_connection, fid);
            if (!f.exists()) {
                throw std::runtime_error("404 Not found");
            }
            if (what == Watch) {
                f.watchers().add(user);
                u.watchedForums().add(fid);
            }
            if (what == Unwatch) {
                f.watchers().remove(user);
                u.watchedForums().remove(fid);
            }
            u.notifiedForums().remove(fid);
        } else {
            throw std::runtime_error("400 Syntax error");
        }
    }
}

/******************************** Commands *******************************/

/* @q USERNEWSRC action:Str [range...] (Talk Command)
   Access user's newsrc file.
   The newsrc file contains read/unread bits for each posting, indexed by posting Id ({@type MID})
   (bit set: post was read; bit clear: post was not read).

   The %action parameter is a keyword that specifies the desired operation:
   - "GET" (get all affected items. Returns a {@type Str|string} of 0/1. In RESP, this doubles as a boolean result.)
   - "SET" (set all affected items (=mark read)).
   - "CLEAR" (clear all affected items (=mark unread)).
   - "ANY" (return 1 if any of the affected items is 1 (=return 1 if any item was read)).
   - "ALL" (return 1 if all of the affected items are 1 (=return 0 if any item was unread)).
   - "FIRSTSET" (return {@type MID} of first set (=read) item, 0 if none).
   - "FIRSTCLEAR" (return {@type MID} of first clear (=unread) item, 0 if none).

   The %range is one or more keyword parameters that specify the items (postings) to check:
   - POST n:{@type MID}... (followed by any number of posting Ids until the end of the command; checks these postings)
   - RANGE a:{@type MID} b:{@type MID} (checks the postings from a (inclusive) to b (not inclusive))
   - THREAD n:{@type TID} (checks all postings in the specified thread)
   - FORUM n:{@type FID} (checks all postings in the specified forum)

   Note that if an error happens, the operation may complete partially.

   Permissions: user context required, accesses user's newsrc

   @err 413 Range error (MID parameter in POST/RANGE is not a valid posting Id)
   @argtype MID
   @argtype TID
   @argtype FID
   @rettype Any
   @rettype MID
   @rettype Str
   @rettype Int
   @uses user:$UID:forum:newsrc:data, user:$UID:forum:newsrc:index */
void
doUserNewsrc(TalkConnection& conn, IntArgBlock& args)
{
    checkUser(conn);

    ReconnectingGuard g(database_rc);
    NewsrcAction action(User(database_connection, conn.getUser()).newsrc());
    int32_t limit = -1;
    while (args.getNumArgs() > 0 && !action.isStopped()) {
        string_t keyword = strUCase(toStringResult(args.getNext()));
        if (keyword == "GET") {
            action.markGet();
        } else if (keyword == "SET") {
            action.markSet();
        } else if (keyword == "CLEAR") {
            action.markClear();
        } else if (keyword == "ANY") {
            action.markFind(true);
        } else if (keyword == "ALL") {
            action.markFind(false);
        } else if (keyword == "FIRSTSET") {
            action.markFind(true);
            action.markWantId();
        } else if (keyword == "FIRSTCLEAR") {
            action.markFind(false);
            action.markWantId();
        } else if (keyword == "POST") {
            // POST n n n -- individual messages
            if (limit == -1) {
                limit = DbSubtree(database_connection, MESSAGE_ROOT).intKey("id").get();
            }
            while (args.getNumArgs() > 0 && !action.isStopped()) {
                int32_t i = toIntResult(args.getNext());
                if (i <= 0 || i > limit) {
                    throw std::runtime_error("413 Range error");
                }
                action.process(i);
            }
        } else if (keyword == "RANGE") {
            // RANGE a b -- interval [a,b)
            if (limit == -1) {
                limit = DbSubtree(database_connection, MESSAGE_ROOT).intKey("id").get();
            }
            args.checkArgumentCountAtLeast(2);
            int32_t a = toIntResult(args.getNext());
            int32_t b = toIntResult(args.getNext());
            if (a <= 0 || b > limit) {
                throw std::runtime_error("413 Range error");
            }
            while (a <= b && !action.isStopped()) {
                action.process(a);
                ++a;
            }
        } else if (keyword == "FORUM") {
            // FORUM -- all messages in a forum
            args.checkArgumentCountAtLeast(1);
            int32_t fid = toIntResult(args.getNext());
            action.process(Forum(database_connection, fid).messages());
        } else if (keyword == "THREAD") {
            // THREAD -- all messages in a thread
            args.checkArgumentCountAtLeast(1);
            int32_t tid = toIntResult(args.getNext());
            action.process(Topic(database_connection, tid).messages());
        } else {
            conn.sendError("400 Syntax error");
            return;
        }
    }
    action.save();
    action.sendResult(conn);
}

/* @q USERWATCH [THREAD n:TID] [FORUM n:FID]... (Talk Command)
   Watch thread or forum.
   Any number of threads or forums can be watched in a single command
   by specifying multiple THREAD or FORUM parameters.

   Permissions: user context required, accesses user's profile

   @err 404 Not found
   @uses user:$UID:forum:watchedThreads, user:$UID:forum:watchedForums
   @uses user:$UID:forum:notifiedThreads, user:$UID:forum:notifiedForums
   @uses forum:$FID:watchers, thread:$TID:watchers
   @see USERLSWATCHEDTHREADS, USERLSWATCHEDFORUMS */
void
doUserWatch(TalkConnection& conn, IntArgBlock& args)
{
    checkUser(conn);
    ReconnectingGuard g(database_rc);
    WatchAction w(WatchAction::Watch);
    w.process(args, conn.getUser());
    conn.sendSuccess("OK");
}

/* @q USERUNWATCH [THREAD n:TID] [FORUM n:FID]... (Talk Command)
   Stop watching thread or forum.
   Any number of threads or forums can be unwatched in a single command
   by specifying multiple THREAD or FORUM parameters.

   Permissions: user context required, accesses user's profile

   @err 404 Not found
   @uses user:$UID:forum:watchedThreads, user:$UID:forum:watchedForums
   @uses user:$UID:forum:notifiedThreads, user:$UID:forum:notifiedForums
   @uses forum:$FID:watchers, thread:$TID:watchers
   @see USERLSWATCHEDTHREADS, USERLSWATCHEDFORUMS */
void
doUserUnwatch(TalkConnection& conn, IntArgBlock& args)
{
    checkUser(conn);
    ReconnectingGuard g(database_rc);
    WatchAction w(WatchAction::Unwatch);
    w.process(args, conn.getUser());
    conn.sendSuccess("OK");
}

/* @q USERMARKSEEN [THREAD n:TID] [FORUM n:FID]... (Talk Command)
   Reset notification status for a thread or forum.
   If a user has set their notifications to "one per/thread forum"
   ({user:$UID:profile}->talkwatchindividual),
   they only get a notification for the first change.
   This command resets the status so a following change will again send mail.

   Any number of threads or forums can be marked seen in a single command
   by specifying multiple THREAD or FORUM parameters.

   Permissions: user context required, accesses user's profile

   @err 404 Not found
   @uses forum:$FID:watchers, thread:$TID:watchers */
void
doUserMarkSeen(TalkConnection& conn, IntArgBlock& args)
{
    checkUser(conn);
    ReconnectingGuard g(database_rc);
    WatchAction w(WatchAction::MarkSeen);
    w.process(args, conn.getUser());
    conn.sendSuccess("OK");
}

/* @q USERLSWATCHEDTHREADS [listParameters...] (Talk Command)
   List threads watched by user.

   The list can be accessed in different ways, see {pcc:talk:listparams|listParameters}.
   Valid sort keys for threads are:
   - firstpost
   - forum
   - lastpost
   - lasttime
   - subject

   Permissions: user context required, accesses user's profile

   @rettype Any
   @rettype TID
   @uses user:$UID:forum:watchedThreads
   @see USERWATCH, USERUNWATCH */
void
doUserListWatchedThreads(TalkConnection& conn, IntArgBlock& args)
{
    // Parse
    checkUser(conn);
    ListParams parms;
    parms.parse(args, false, ListParams::threadSortKeys);

    // Do it
    ReconnectingGuard g(database_rc);
    User u(database_connection, conn.getUser());
    parms.exec(conn, u.watchedTopics());
}

/* @q USERLSWATCHEDFORUMS [listParameters...] (Talk Command)
   List forums watched by user.

   The list can be accessed in different ways, see {pcc:talk:listparams|listParameters}.
   Valid sort keys for forums are:
   - key
   - lastpost
   - lasttime
   - name

   Permissions: user context required, accesses user's profile

   @rettype Any
   @rettype FID
   @uses user:$UID:forum:watchedForums
   @see USERWATCH, USERUNWATCH */
void
doUserListWatchedForums(TalkConnection& conn, IntArgBlock& args)
{
    // Parse
    checkUser(conn);
    ListParams parms;
    parms.parse(args, false, ListParams::forumSortKeys);

    // Do it
    ReconnectingGuard g(database_rc);
    User u(database_connection, conn.getUser());
    parms.exec(conn, u.watchedForums());
}

/* @q USERLSPOSTED user:UID [listParameters...] (Talk Command)
   List user's postings.

   The list can be accessed in different ways, see {pcc:talk:listparams|listParameters}.
   Valid sort keys for postings are:
   - author
   - edittime
   - subject
   - thread
   - time

   Permissions: none (everyone can execute this command).

   @rettype Any
   @rettype MID
   @uses user:$UID:forum:posted */
void
doUserListPosted(TalkConnection& conn, IntArgBlock& args)
{
    // Parse
    args.checkArgumentCountAtLeast(1);
    string_t user(toStringResult(args.getNext()));

    ListParams parms;
    parms.parse(args, false, ListParams::postSortKeys);

    // Do it
    ReconnectingGuard g(database_rc);
    User u(database_connection, user);
    parms.exec(conn, u.postedMessages());
}
