/**
  *  \file planetscentral/userdir.cc
  *  \brief File Server - User Directory Objects
  */

#include <iostream>
#include "planetscentral/userdir.h"
#include "io/dirs.h"
#include "io/textfile.h"

static const char CONTROL_FILE[] = ".c2file";

static int
permsFromString(string_t s)
{
    int result = 0;
    if (s.find('r') != s.npos)
        result += UserDirectory::AllowRead;
    if (s.find('w') != s.npos)
        result += UserDirectory::AllowWrite;
    if (s.find('l') != s.npos)
        result += UserDirectory::AllowList;
    if (s.find('a') != s.npos)
        result += UserDirectory::AllowAccess;
    return result;
}

static string_t
permsToString(int n)
{
    string_t result;
    if (n & UserDirectory::AllowRead)
        result += 'r';
    if (n & UserDirectory::AllowWrite)
        result += 'w';
    if (n & UserDirectory::AllowList)
        result += 'l';
    if (n & UserDirectory::AllowAccess)
        result += 'a';
    if (result.empty())
        result += '0';
    return result;
}

/******************************** UserItem *******************************/

/** Constructor.
    \param name Name of this object (basename)
    \param parent Parent directory */
UserItem::UserItem(string_t name, UserDirectory* parent)
    : name(name),
      parent(parent)
{ }

UserItem::~UserItem()
{ }

string_t
UserItem::getName() const
{
    return name;
}

UserDirectory*
UserItem::getParent() const
{
    return parent;
}

/******************************** UserFile *******************************/

/** Constructor.
    \param name Name of this file (basename)
    \param parent Parent directory */
UserFile::UserFile(string_t name, UserDirectory* parent)
    : UserItem(name, parent),
      size(0)
{ }

UserFile::~UserFile()
{ }

/** Set file meta-information. Right now, we only store the size.
    \param size Size in bytes */
void
UserFile::setInfo(uint32_t size)
{
    this->size = size;
}

/** Get file size. */
uint32_t
UserFile::getSize() const
{
    return size;
}

/***************************** UserDirectory *****************************/

/** Constructor.
    \param name Name of this directory (basename)
    \param parent Parent directory */
UserDirectory::UserDirectory(string_t name, UserDirectory* parent)
    : UserItem(name, parent),
      was_read(false),
      num_fully_read_subdirs(0),
      dir_info(0)
{ }

UserDirectory::~UserDirectory()
{
    std::cout << "discarding " << getFullPath() << std::endl;
}

/** Check whether this directory was read. Only when this is true, most
    accessors can be used. */
bool
UserDirectory::wasRead() const
{
    return was_read;
}

/** Check whether this directory was fully read. A directory is fully
    read if all its subdirectories are fully read. */
bool
UserDirectory::wasFullyRead() const
{
    return was_read && num_fully_read_subdirs == subdirs.size();
}

/** Ensure directory was read.
    \param physical Physical directory to read */
void
UserDirectory::read(string_t physical)
{
    if (was_read)
        return;
    was_read = true;

    /* Read directory */
    std::cout << "reading " << physical << " ('" << getName() << "')" << std::endl;
    try {
        /* Read directory entries */
        Ptr<Directory> dir = Directory::openDir(physical);
        Ptr<Enumeration<Ptr<DirectoryEntry> > > entries = dir->getEntries();
        while (entries->hasMoreElements()) {
            Ptr<DirectoryEntry> e = entries->getNextElement();
            string_t name = e->getBaseName();
            if (name.size() == 0 || name[0] == '.') {
                /* skip */
            } else if (e->getFileType() == DirectoryEntry::FT_FILE) {
                /* file */
                files.push_back(new UserFile(name, this));
                files.back()->setInfo(e->getSize());
            } else if (e->getFileType() == DirectoryEntry::FT_DIRECTORY) {
                /* directory */
                subdirs.push_back(new UserDirectory(name, this));
            }
        }

        /* Read metainfo */
        loadControlFile(*dir);
        setOwner();

        if (wasFullyRead() && getParent() != 0) {
            ++getParent()->num_fully_read_subdirs;
        }
    }
    catch (std::exception& e) {
        std::cout << "   error: " << e.what() << std::endl;
    }
}

/** Ensure directory was read. */
void
UserDirectory::read()
{
    if (!was_read)
        read(getFullPath());
}

/** Ensure directory was read fully.
    \param physical Physical directory to read */
void
UserDirectory::readFully(string_t physical)
{
    if (wasFullyRead())
        return;

    read(physical);
    for (ptr_vector<UserDirectory>::size_type i = 0; i < subdirs.size(); ++i) {
        if (!subdirs[i]->wasFullyRead()) {
            subdirs[i]->readFully(Directory::makeFileName(physical, subdirs[i]->getName()));
        }
    }

    if (wasFullyRead() && getParent() != 0) {
        ++getParent()->num_fully_read_subdirs;
    }
}

/** Ensure directory was read fully. */
void
UserDirectory::readFully()
{
    readFully(getFullPath());
}

/** Forget this directory and all its subdirectories. It returns to
    not-read status. */
void
UserDirectory::forget()
{
    if (was_read) {
        std::cout << "forgetting " << getFullPath() << std::endl;
        if (wasFullyRead() && getParent() != 0) {
            --getParent()->num_fully_read_subdirs;
        }
        subdirs.clear();
        files.clear();
        dir_info.reset();
        control_info.clear();
        owner.clear();
        was_read = false;
        num_fully_read_subdirs = 0;
    }
}

/** Get full physical path name for this directory. */
string_t
UserDirectory::getFullPath()
{
    return Directory::makeFileName(getParent()->getFullPath(), getName());
}

/** Find item in this directory.
    \param name Name of item to find
    \pre wasRead()
    \return Item if found, null if not found */
UserItem*
UserDirectory::find(string_t name)
{
    ASSERT(wasRead());
    for (ptr_vector<UserDirectory>::size_type i = 0; i < subdirs.size(); ++i) {
        if (name == subdirs[i]->getName()) {
            return subdirs[i];
        }
    }
    for (ptr_vector<UserFile>::size_type i = 0; i < files.size(); ++i) {
        if (name == files[i]->getName()) {
            return files[i];
        }
    }
    return 0;
}

/** Get number of subdirectories.
    \pre wasRead() */
uint32_t
UserDirectory::getNumDirs() const
{
    ASSERT(wasRead());
    return subdirs.size();
}

/** Get number of files.
    \pre wasRead() */
uint32_t
UserDirectory::getNumFiles() const
{
    ASSERT(wasRead());
    return files.size();
}

/** Get directory by index. */
UserDirectory*
UserDirectory::getDir(uint32_t n)
{
    ASSERT(wasRead());
    ASSERT(n < subdirs.size());
    return subdirs[n];
}

/** Get file by index. */
UserFile*
UserDirectory::getFile(uint32_t n)
{
    ASSERT(wasRead());
    ASSERT(n < files.size());
    return files[n];
}

/** Get owner of this directory. */
string_t
UserDirectory::getOwner() const
{
    ASSERT(wasRead());
    return owner;
}

/** Get permissions for a user.
    \param user User Id */
int
UserDirectory::getPermissions(string_t user) const
{
    ASSERT(wasRead());
    std::map<string_t,string_t>::const_iterator i;
    if (user.empty() || user == getOwner()) {
        /* Owner and root can do anything */
        return AllowAll;
    } else if ((i = control_info.find("perms:" + user)) != control_info.end()) {
        /* Explicit permissions for this user */
        return permsFromString(i->second);
    } else if ((i = control_info.find("perms:*")) != control_info.end()) {
        /* Wildcard permissions */
        return permsFromString(i->second);
    } else {
        /* Refuse anyone else */
        return 0;
    }
}

/** Set user permissions. */
void
UserDirectory::setPermissions(string_t user, string_t perms)
{
    ASSERT(wasRead());
    setProperty("perms:" + user, permsToString(permsFromString(perms)));
}

/** List all permissions. */
void
UserDirectory::listPermissions(std::vector<std::pair<string_t, string_t> >& perms)
{
    ASSERT(wasRead());
    for (std::map<string_t,string_t>::const_iterator i = control_info.begin(), e = control_info.end(); i != e; ++i) {
        if (i->first.size() > 6 && i->first.compare(0, 6, "perms:", 6) == 0) {
            perms.push_back(std::make_pair(i->first.substr(6), i->second));
        }
    }
}

/** Get visibility level for a directory.
    \retval 0 only owner can access it
    \retval 1 some permissions have been granted
    \retval 2 world-permissions have been granted */
int
UserDirectory::getVisibilityLevel() const
{
    ASSERT(wasRead());
    for (std::map<string_t,string_t>::const_iterator i = control_info.begin(), e = control_info.end(); i != e; ++i) {
        if (i->first.size() > 6 && i->first.compare(0, 6, "perms:", 6) == 0) {
            // It's a permission entry.
            if (permsFromString(i->second) != 0) {
                if (i->first == "perms:*") {
                    return 2;
                } else {
                    return 1;
                }
            }
        }
    }
    return 0;
}


/** Get directory property. This takes full property names, and can
    thus be used to modify all properties of a directory. Properties
    modified with PROPSET have a "prop:" prefix.
    \param name Name of property */
string_t
UserDirectory::getProperty(string_t name) const
{
    ASSERT(wasRead());
    std::map<string_t,string_t>::const_iterator i = control_info.find(name);
    if (i != control_info.end()) {
        return i->second;
    } else {
        return string_t();
    }
}

/** Modify directory property. This takes full property names, and can
    thus be used to modify all properties of a directory. Properties
    modified with PROPSET have a "prop:" prefix.

    Note that the name must not contain "=", "\r" or "\n"; the value
    must not contain "\r" or "\n".

    \param name Name of property
    \param value Value of property */
void
UserDirectory::setProperty(string_t name, string_t value)
{
    ASSERT(wasRead());
    control_info[name] = value;
    saveControlFile();
    if (name == "owner")
        setOwner();
}

/** Get directory information. The information will be read if it
    is not yet present. */
DirInfo&
UserDirectory::getDirInfo()
{
    if (!dir_info.get()) {
        string_t name = getFullPath();
        std::cout << "checking " << name << std::endl;
        Ptr<Directory> dir = Directory::openDir(name);
        dir_info.reset(new DirInfo(*dir));
    }
    return *dir_info;
}

/** Create subdirectory. This function automatically reads the directory
    if required.
    \param name Name of subdirectory
    \param forUser Create directory owned by this user. May be blank
    to create a normal directory which inherits the user.
    \retval true Success
    \retval false Error */
bool
UserDirectory::createDirectory(string_t name, string_t forUser)
{
    read();
    try {
        /* Create directory */
        Ptr<Directory> dir = Directory::openDir(getFullPath());
        dir->getEntryByName(name)->createAsDirectory();

        /* Add to list */
        UserDirectory* d = new UserDirectory(name, this);
        subdirs.push_back(d);

        /* Create control file */
        if (forUser.size()) {
            /* setProperty implies setOwner and saveControlFile,
               but requires read() */
            d->read();
            d->setProperty("owner", forUser);
        }
        return true;
    }
    catch (...) {
        return false;
    }
}

/** Create file. This function automatically reads the directory
    if required.
    \param name Name of file
    \param data Data to store in file
    \retval true Success
    \retval false Error */
bool
UserDirectory::createFile(string_t name, const string_t& data)
{
    /* Check for existing item */
    read();
    dir_info.reset();
    UserItem* item = find(name);
    UserFile* file = dynamic_cast<UserFile*>(item);
    if (item != 0 && file == 0) {
        /* exists, but is not a file */
        std::cout << "cannot create file '" << name << "' because it exists as a non-file" << std::endl;
        return false;
    }

    /* Write it */
    try {
        Ptr<Stream> s = openFile(name, Stream::C_CREATE);
        if (!s) {
            std::cout << "creation of file '" << name << "' failed" << std::endl;
            return false;
        }
        s->writeT(data.data(), data.size());
        if (!file) {
            file = new UserFile(name, this);
            files.push_back(file);
        }
        file->setInfo(data.size());
        return true;
    }
    catch (...) {
        std::cout << "error while creating file '" << name << "'" << std::endl;
        return false;
    }
}

/** Remove entry.
    \param it Item to remove
    \return true on success
    \todo This is a very simple implementation. A more clever implementation
    would remove \c it from \c subdirs or \c files instead of forgetting the
    whole directory. */
bool
UserDirectory::removeEntry(UserItem* it)
{
    /* No need to read, because /it/ can only be valid if we were
       already read. */
    if (/*UserFile* uf =*/ dynamic_cast<UserFile*>(it)) {
        /* It's a file. This is easy. */
        Ptr<Directory> dir = Directory::openDir(getFullPath());
        if (dir->eraseNT(it->getName())) {
            forget();
            dir_info.reset();
            return true;
        } else {
            return false;
        }
    } else if (UserDirectory* ud = dynamic_cast<UserDirectory*>(it)) {
        /* It's a directory. It might not be empty. Even if it is
           empty, it might contain a control file. */
        ud->read();
        if (ud->getNumDirs() != 0 || ud->getNumFiles() != 0) {
            return false;
        } else {
            Ptr<Directory> dir = Directory::openDir(getFullPath());
            Ptr<Directory> sub = dir->openSubdir(ud->getName());
            sub->eraseNT(CONTROL_FILE);
            if (dir->eraseNT(ud->getName())) {
                forget();
                dir_info.reset();
                return true;
            } else {
                return false;
            }
        }
    } else {
        return false;
    }
}

/** Remove all entries of this directory. */
bool
UserDirectory::removeAllEntries()
{
    ASSERT(wasRead());
    Ptr<Directory> dir = Directory::openDir(getFullPath());

    /* Files */
    bool ok = true;
    for (uint32_t i = 0; i < files.size() && ok; ++i) {
        ok = dir->eraseNT(files[i]->getName());
    }

    /* Directories */
    for (uint32_t i = 0; i < subdirs.size() && ok; ++i) {
        UserDirectory* ud = subdirs[i];
        if (ud->wasRead() && (ud->getNumDirs() != 0 || ud->getNumFiles() != 0)) {
            ok = false;
        } else {
            Ptr<Directory> sub = dir->openSubdir(ud->getName());
            sub->eraseNT(CONTROL_FILE);
            ok = dir->eraseNT(ud->getName());
        }
    }

    /* In any case, forget this directory */
    forget();
    dir_info.reset();
    return ok;
}

/** Open file.
    \param name Name of file
    \param mode Mode (see Stream::openFile).
    \return opened stream, null on error */
Ptr<Stream>
UserDirectory::openFile(string_t name, int mode)
{
    Ptr<Directory> dir = Directory::openDir(getFullPath());
    return dir->openFileNT(name, mode);
}

/** Compute directory totals.
    \param nfiles [out] Number of files added here
    \param nkbytes [out] Size in kilobytes added here
    \pre wasFullyRead() */
void
UserDirectory::computeTotals(int32_t& nfiles, int32_t& nkbytes)
{
    /* Count ourselves */
    ++nfiles;
    ++nkbytes;

    /* Count content */
    for (ptr_vector<UserDirectory>::size_type i = 0; i < subdirs.size(); ++i) {
        subdirs[i]->computeTotals(nfiles, nkbytes);
    }
    for (ptr_vector<UserFile>::size_type i = 0; i < files.size(); ++i) {
        ++nfiles;
        nkbytes += (files[i]->getSize() + 1023) / 1024;
    }
}

/** Load control file.
    \param dir Directory to read from */
void
UserDirectory::loadControlFile(Directory& dir)
{
    Ptr<Stream> config = dir.openFileNT(CONTROL_FILE, Stream::C_READ);
    if (config) {
        TextFile tf(*config);
        tf.setUtf8();
        string_t line;
        while (tf.getLine(line)) {
            string_t::size_type n = line.find('=');
            if (n != line.npos) {
                control_info.insert(std::make_pair(string_t(line, 0, n), string_t(line, n+1)));
            }
        }
    }
}

/** Save control file.
    \param dir Directory to save in */
void
UserDirectory::saveControlFile(Directory& dir)
{
    if (control_info.empty()) {
        dir.eraseNT(CONTROL_FILE);
    } else {
        Ptr<Stream> config = dir.openFile(CONTROL_FILE, Stream::C_CREATE);
        TextFile tf(*config);
        tf.setUtf8();
        for (std::map<string_t,string_t>::const_iterator i = control_info.begin(); i != control_info.end(); ++i) {
            tf.writeLine(i->first + "=" + i->second);
        }
        tf.sync();
    }
}

/** Save control file. Automatically determines the directory to save in. */
void
UserDirectory::saveControlFile()
{
    Ptr<Directory> dir = Directory::openDir(getFullPath());
    saveControlFile(*dir);
}

/** Set directory owner from control file. If the control file
    specifies an owner, use that; otherwise, use the parent's. */
void
UserDirectory::setOwner()
{
    std::map<string_t,string_t>::const_iterator i = control_info.find("owner");
    if (i != control_info.end()) {
        owner = i->second;
    } else if (getParent()) {
        owner = getParent()->getOwner();
    } else {
        owner.clear();
    }
}

/******************************** UserRoot *******************************/

/** Root directory.
    \param name Display name
    \param physical Physical directory name */
UserRoot::UserRoot(string_t name, string_t physical)
    : UserDirectory(name, 0),
      physical(physical)
{ }

UserRoot::~UserRoot()
{ }

string_t
UserRoot::getFullPath()
{
    return physical;
}
